import numpy as np
import scipy.optimize as opt
edge_p = 0.05
edge_sample = 1.0
# node_sample = 0.9
dimension = 7
node_num = 1500


def groundTruth():
    G = np.zeros((node_num, node_num))
    for i in range(node_num):
        G[i][i] = 1
        for j in range(i+1, node_num):
            G[i][j] = np.random.choice(range(2), p=[1-edge_p, edge_p])
            G[j][i] = G[i][j]
    feature_g = np.zeros((node_num, dimension))
    for i in range(node_num):
        for j in range(dimension):
            feature_g[i][j] = np.random.choice(range(2), 1)
    return G, feature_g


def dataGenerator(G, feature_g, isMain, node_sample):
    if isMain:
        A = np.zeros((node_num, node_num))
        for i in range(node_num):
            A[i][i] = 1
            for j in range(i+1, node_num):
                A[i][j] = np.random.choice([G[i][j], 0], p=[edge_sample, 1 - edge_sample])
                A[j][i] = A[i][j]
        P = np.identity(node_num)
        feature_a = np.zeros((node_num, dimension))
        for i in range(node_num):
            for j in range(dimension):
                feature_a[i][j] = feature_g[i][j]
        return A, feature_a, P

    else:
        b = np.zeros((node_num, node_num))
        for i in range(node_num):
            b[i][i] = 1
            for j in range(i+1, node_num):
                b[i][j] = np.random.choice([G[i][j], 0], p=[edge_sample, 1 - edge_sample])
                b[j][i] = b[i][j]

        node_g = np.zeros((1, node_num))
        available = []
        count = 0
        for i in range(node_num):
            available.append(i)
            node_g[0][i] = i
            isExist = np.random.choice(range(2), p=[1 - node_sample, node_sample])
            if isExist == 1:
                count += 1

        transform = np.zeros((node_num, count))
        for i in range(count):
            j = np.random.choice(available, 1)
            available.remove(j)
            # G/A中的j是B中的i
            transform[j[0]][i] = 1
        # print("this is transform", transform)

        B = np.matmul(np.matmul(transform.T, b), transform)
        feature_b = np.zeros((count, dimension))
        row, col = opt.linear_sum_assignment(-transform.T)
        # print(row, "r")
        # print(col, "c")
        for i in range(count):
            for j in range(dimension):
                feature_b[i][j] = feature_g[col[i]][j]
        return B, feature_b, transform


def loss(matrix_a, matrix_b, matrix_p, matrix_d, beta):
    tmp_1 = matrix_a - np.matmul(np.matmul(matrix_p, matrix_b), matrix_p.T)
    # print("tmp1_max", tmp_1.max())
    # tmp_1 = tmp_1/tmp_1.max()
    tmp_2 = matrix_p * matrix_d
    # print("tmp2_max", tmp_2.max())
    # tmp_2 = tmp_2/tmp_2.max()
    error = np.linalg.norm(tmp_1) ** 2 + beta * (np.linalg.norm(tmp_2) ** 2)
    return error


def distance(feature_a, feature_b):
    a = len(feature_a)
    b = len(feature_b)
    D = np.zeros((a, b))
    for i in range(a):
        for j in range(b):
            for k in range(dimension):
                D[i][j] += (feature_a[i][k] - feature_b[j][k]) ** 2
    return D


def training(matrix_a, matrix_b, matrix_d, beta):
    a = len(matrix_a)
    b = len(matrix_b)
    if a > b:
        c = b
    else:
        c = a
    matrix_p = np.ones((a, b))/c
    threshold = 0.001
    error = loss(matrix_a, matrix_b, matrix_p, matrix_d, beta)
    counter = 0
    while error > threshold:
        counter += 1
        if counter > 100:
            break
        old = error
        gradient = 4 * np.matmul(np.matmul(matrix_a, matrix_p), matrix_b)
        gradient -= 2 * beta * np.linalg.norm(matrix_p * matrix_d) * matrix_d
        row, col = opt.linear_sum_assignment(-gradient)
        matrix_q = np.zeros((a, b))
        for y in range(c):
            matrix_q[row[y]][col[y]] = 1
        matrix_p = 0.1 * matrix_p + 0.9 * matrix_q
        # print(counter, matrix_p)
        error = loss(matrix_a, matrix_b, matrix_p, matrix_d, beta)
        delta = old - error
        # print(counter, error, delta)
    row, col = opt.linear_sum_assignment(-matrix_p)
    matrix_p = np.zeros((a, b))
    for y in range(c):
        matrix_p[row[y]][col[y]] = 1
    return matrix_p


def sinkhorn(matrix_p):
    a = len(matrix_p)
    b = len(matrix_p.T)
    matrix_q = np.zeros((a, b))
    epsilon = 0.0001
    error = np.linalg.norm(matrix_p - matrix_q) ** 2
    while error > epsilon:
        for j in range(b):
            count = np.sum(matrix_p.T[j])
            for i in range(a):
                matrix_q[i][j] = matrix_p/count
        for i in range(a):
            count = np.sum(matrix_q[i])
            for j in range(b):
                matrix_p[i][j] = matrix_p/count
        error = np.linalg.norm(matrix_p - matrix_q) ** 2
    return matrix_q


def accuracy(P, P_algorithm):
    a = len(P)
    b = len(P.T)
    count = 0
    for i in range(a):
        for j in range(b):
            if P[i][j] == 1:
                if P_algorithm[i][j] == 1:
                    count += 1
    total = np.linalg.norm(P) ** 2
    accurate = count / total
    print("accuracy:", accurate)
    return


if __name__ == '__main__':
    G, feature_g = groundTruth()
    # print("G", G)
    # print("feature_g", feature_g)

    B, feature_bb, transform1 = dataGenerator(G, feature_g, False, 0.9)
    # print("B", B, len(B))
    # print("feature_b", feature_bb)
    # print("transform", transform1)

    C, feature_cc, transform2 = dataGenerator(G, feature_g, False, 0.2)
    # print("C", C, len(C))
    # print("feature_c", feature_cc)
    # print("transform", transform2)

    matD = distance(feature_bb, feature_cc)
    '''
    matD1 = distance(feature_g, feature_bb)
    matD2 = distance(feature_g, feature_cc)
    alg1 = training(G, B, matD1, 0)
    alg2 = training(G, C, matD2, 0)
    alg5 = training(G, B, matD1, 1.0)
    alg6 = training(G, C, matD2, 1.0)
    '''
    alg3 = training(B, C, matD, 0)
    alg4 = training(B, C, matD, 1.0)
    '''
    alg7 = training(B, C, matD, 0.0938)
    alg8 = training(B, C, matD, 2.0)
    '''
    '''
    print("B, 0")
    accuracy(transform1, alg1)
    print("B, 1.0")
    accuracy(transform1, alg5)
    print("C, 0")
    accuracy(transform2, alg2)
    print("C, 1.0")
    accuracy(transform2, alg6)
    '''
    print("BC 0")
    accuracy(np.matmul(transform1.T, transform2), alg3)
    print("BC 1")
    accuracy(np.matmul(transform1.T, transform2), alg4)
    '''
    print("BC 0.5")
    accuracy(np.matmul(transform1.T, transform2), alg7)
    print("BC 2")
    accuracy(np.matmul(transform1.T, transform2), alg8)
    '''
